@@ -27,6 +27,8 @@ from ._events import ResponseEndOfMessage
 from ._events import ResponseHeaders
 from ._events import ResponseProtocolError
 from ._events import ResponseTrailers
+from ._hooks import HttpConnectedHook
+from ._hooks import HttpConnectErrorHook
 from ._hooks import HttpConnectHook
 from ._hooks import HttpErrorHook
 from ._hooks import HttpRequestHeadersHook
@@ -59,6 +61,7 @@ from mitmproxy.proxy.layers import tls
 from mitmproxy.proxy.layers import websocket
 from mitmproxy.proxy.layers.http import _upstream_proxy
 from mitmproxy.proxy.utils import expect
+from mitmproxy.proxy.utils import ReceiveBuffer
 from mitmproxy.utils import human
 from mitmproxy.websocket import WebSocketData
 
@@ -143,8 +146,8 @@ class DropStream(HttpCommand):
 
 
 class HttpStream(layer.Layer):
-    request_body_buf: bytes
-    response_body_buf: bytes
+    request_body_buf: ReceiveBuffer
+    response_body_buf: ReceiveBuffer
     flow: http.HTTPFlow
     stream_id: StreamId
     child_layer: layer.Layer | None = None
@@ -158,8 +161,8 @@ class HttpStream(layer.Layer):
 
     def __init__(self, context: Context, stream_id: int) -> None:
         super().__init__(context)
-        self.request_body_buf = b""
-        self.response_body_buf = b""
+        self.request_body_buf = ReceiveBuffer()
+        self.response_body_buf = ReceiveBuffer()
         self.client_state = self.state_uninitialized
         self.server_state = self.state_uninitialized
         self.stream_id = stream_id
@@ -352,8 +355,8 @@ class HttpStream(layer.Layer):
             self.flow.request.trailers = event.trailers
         elif isinstance(event, RequestEndOfMessage):
             self.flow.request.timestamp_end = time.time()
-            self.flow.request.data.content = self.request_body_buf
-            self.request_body_buf = b""
+            self.flow.request.data.content = bytes(self.request_body_buf)
+            self.request_body_buf.clear()
             self.client_state = self.state_done
             yield HttpRequestHook(self.flow)
             if (yield from self.check_killed(True)):
@@ -459,8 +462,8 @@ class HttpStream(layer.Layer):
             self.flow.response.trailers = event.trailers
         elif isinstance(event, ResponseEndOfMessage):
             assert self.flow.response
-            self.flow.response.data.content = self.response_body_buf
-            self.response_body_buf = b""
+            self.flow.response.data.content = bytes(self.response_body_buf)
+            self.response_body_buf.clear()
             yield from self.send_response()
 
     def send_response(self, already_streamed: bool = False):
@@ -605,16 +608,16 @@ class HttpStream(layer.Layer):
                 self.flow.request.stream = True
                 if self.request_body_buf:
                     # clear buffer and then fake a DataReceived event with everything we had in the buffer so far.
-                    body_buf = self.request_body_buf
-                    self.request_body_buf = b""
+                    body_buf = bytes(self.request_body_buf)
+                    self.request_body_buf.clear()
                     yield from self.start_request_stream()
                     yield from self.handle_event(RequestData(self.stream_id, body_buf))
             if response:
                 assert self.flow.response
                 self.flow.response.stream = True
                 if self.response_body_buf:
-                    body_buf = self.response_body_buf
-                    self.response_body_buf = b""
+                    body_buf = bytes(self.response_body_buf)
+                    self.response_body_buf.clear()
                     yield from self.start_response_stream()
                     yield from self.handle_event(ResponseData(self.stream_id, body_buf))
         return False
@@ -748,16 +751,30 @@ class HttpStream(layer.Layer):
             )
 
         if 200 <= self.flow.response.status_code < 300:
+            yield HttpConnectedHook(self.flow)
             self.child_layer = self.child_layer or layer.NextLayer(self.context)
             self._handle_event = self.passthrough
             yield from self.child_layer.handle_event(events.Start())
+        else:
+            yield HttpConnectErrorHook(self.flow)
+            self.client_state = self.state_errored
+            self.flow.live = False
+
+        content = self.flow.response.raw_content
+        done_after_headers = not (content or self.flow.response.trailers)
+        yield SendHttp(
+            ResponseHeaders(self.stream_id, self.flow.response, done_after_headers),
+            self.context.client,
+        )
+        if content:
+            yield SendHttp(ResponseData(self.stream_id, content), self.context.client)
+
+        if self.flow.response.trailers:
             yield SendHttp(
-                ResponseHeaders(self.stream_id, self.flow.response, True),
+                ResponseTrailers(self.stream_id, self.flow.response.trailers),
                 self.context.client,
             )
-            yield SendHttp(ResponseEndOfMessage(self.stream_id), self.context.client)
-        else:
-            yield from self.send_response()
+        yield SendHttp(ResponseEndOfMessage(self.stream_id), self.context.client)
 
     @expect(RequestData, RequestEndOfMessage, events.Event)
     def passthrough(self, event: events.Event) -> layer.CommandGenerator[None]:
