@@ -21,13 +21,11 @@ from mitmproxy.utils import strutils
 
 
 @overload
-def read_file(filename: str, escaped: Literal[True]) -> bytes:
-    ...
+def read_file(filename: str, escaped: Literal[True]) -> bytes: ...
 
 
 @overload
-def read_file(filename: str, escaped: Literal[False]) -> str:
-    ...
+def read_file(filename: str, escaped: Literal[False]) -> str: ...
 
 
 def read_file(filename: str, escaped: bool) -> bytes | str:
